/* Revision: 2.8.5 */

/******************************************************************************
* Copyright 1998-2017 NetBurner, Inc.  ALL RIGHTS RESERVED
*
*    Permission is hereby granted to purchasers of NetBurner Hardware to use or
*    modify this computer program for any use as long as the resultant program
*    is only executed on NetBurner provided hardware.
*
*    No other rights to use this program or its derivatives in part or in
*    whole are granted.
*
*    It may be possible to license this or other NetBurner software for use on
*    non-NetBurner Hardware. Contact sales@Netburner.com for more information.
*
*    NetBurner makes no representation or warranties with respect to the
*    performance of this computer program, and specifically disclaims any
*    responsibility for any damages, special or consequential, connected with
*    the use of this program.
*
* NetBurner
* 5405 Morehouse Dr.
* San Diego, CA 92121
* www.netburner.com
******************************************************************************/

/* NB Library Definitions */
#include "predef.h"

/* C Standard Library */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Portability & uCos Definitions */
#include <includes.h>

/* NB Runtime Libraries */
#include <constants.h>
#include <utils.h>
#include <buffers.h>
#include <system.h>

/* NB Network Stack */
#include <arp.h>
#include <dhcpd.h>
#include <ip.h>
#include <udp.h>
#include <tcp.h>
#include <counters.h>
#include <nettimer.h>
#include <netinterface.h>
#include <ethernet.h>

/* SNMP */
#include <snmp.h>

#ifdef _DEBUG
   #define C_CSUM
#endif

#if defined MULTIHOME || defined AUTOIP
   #include <multihome.h>
#endif

VWORD wGlobalPktId FAST_IP_VAR; /* The counter for packets */
BYTE bTTL_Default FAST_IP_VAR;
OS_FIFO ipRxFifo FAST_IP_VAR;
OS_FIFO *IcmpFifo FAST_IP_VAR;

SecondTickFunction *pDHCPTick FAST_IP_VAR;
SecondTickFunction *pDHCPServerTick FAST_IP_VAR;
SecondTickFunction *pMultiTick FAST_IP_VAR;
#ifdef AUTOIP
SecondTickFunction *pAutoIPTick FAST_IP_VAR;
#endif

#ifdef IPV6
#include <ipv6/ipv6_intf.h>
SecondTickFunction *pIpv6Tick FAST_IP_VAR;
#endif



PromisciousPacketFunc    *pPromisciousPacketFunc FAST_IP_VAR;
MULTICastProcessFunction *pMultiCastFunc FAST_IP_VAR;
FragmentProcessFunction  *pFragFunc FAST_IP_VAR;

uint16_t i50msCountEthernet;



#if defined MULTIHOME || defined AUTOIP

int RegisterMultiHomeInterface( InterfaceBlock *pIfBlock );
int nAdditional_Homes FAST_IP_VAR;
int multi_home_if_record[NUM_MULTI_INTERFACES] FAST_IP_VAR;
InterfaceBlock multi_if_block[NUM_MULTI_INTERFACES] FAST_IP_VAR;



int AddInterface(IPADDR4 addr, IPADDR4 mask,IPADDR4 gateway, int root_if)
{
    if (nAdditional_Homes >= NUM_MULTI_INTERFACES) return -1;
    if (root_if==0) root_if=GetFirstInterface();
    InterfaceBlock * cBlock= GetInterFaceBlock(root_if);
    if (cBlock==NULL) return -1;

    multi_if_block[nAdditional_Homes]=*cBlock;
    multi_if_block[nAdditional_Homes].netIP=addr;
    multi_if_block[nAdditional_Homes].netIpMask=mask;
    multi_if_block[nAdditional_Homes].netIpGate=gateway;
    multi_if_block[nAdditional_Homes].bMultiHome=TRUE;
    multi_if_block[nAdditional_Homes].root_if=root_if;
    multi_if_block[nAdditional_Homes].linkActiveFuncPtr = NULL;
    multi_if_block[nAdditional_Homes].pChild = NULL;
    multi_if_block[nAdditional_Homes].dhcpClient = NULL;
    multi_if_block[nAdditional_Homes].AutoIPC = NULL;


    cBlock->pChild = multi_if_block +nAdditional_Homes;

    int ifn=RegisterMultiHomeInterface(&(multi_if_block[nAdditional_Homes]));

    multi_home_if_record[nAdditional_Homes]=ifn;
    nAdditional_Homes++;
    return ifn;
}

int GetMultiHomeInterface4(IPADDR4 ipa, int norgif)
{
    for (int i=0; i< nAdditional_Homes; i++)
    {
        if ((InterfaceIP(multi_home_if_record[i])==ipa) && (norgif==multi_if_block[i].root_if))
        {
            return multi_home_if_record[i];
        }
    }

    return norgif;
}

int GetLocalIPInterface4(IPADDR4 ipa)
{
    for (int i=0; i< nAdditional_Homes; i++)
    {
        if (InterfaceIP(multi_home_if_record[i])==ipa) return multi_home_if_record[i];
    }
    return GetFirstInterface();
}
#endif










void DoIPPacket( PoolPtr pp, PEFRAME pF, WORD csum );



//Make sure they're 4 byte aligned to keep the ColdFire happy
DWORD   IPStk[IP_STK_SIZE] __attribute__( ( aligned( 4 ) ) ) FAST_IP_STK;


void IPTask( void *p );


/*---------------------------------------**
** Check to see if it is a local address **
**---------------------------------------*/
BOOL IsLocal4( IPADDR4 ip, int ifc )
{
   InterfaceBlock *pifb = GetInterFaceBlock( ifc );
   if ( pifb )
   {
   if (pifb->netIP.IsNull()) return FALSE;
   return ( ( ip & pifb->netIpMask ) == ( pifb->netIP & pifb->netIpMask ) );
   }
   return FALSE;
}

/*------------------------------------------**
** Check to see if the address is broadcast **
**------------------------------------------*/
BOOL IsBroadCast( IPADDR4 ip, int ifc )
{
   if ( ip.IsGlobalBroadCast() )
   {
      return TRUE;
   }


   InterfaceBlock *pifb = GetInterFaceBlock( ifc );

   if ( pifb )
   {
      return ip.IsBcastNetMask(pifb->netIP, pifb->netIpMask);
   }
   return FALSE;
}




#ifndef C_CSUM
/*  WORD GetSum( PWORD addr, WORD count )  */
asm(".text");
asm(".global GetSum");
asm("GetSum:");
asm("   moveal %sp@(4),%a0 /* Pointer to data */");
asm("   clr.l %d1");
asm("   movel %d2, %sp@- |save %d2");
asm("   clr.l %d2");
asm("   movel %sp@(12),%d0 /* count */");
asm("   cmpi.l #48, %d0");
asm("   blt Sum_Dword");
asm("   suba.l #16, %sp");
asm("    movem.l #0x0078, %sp@  | save d3-d6 to stack");
asm(" Quad_Dword_Loop:");
asm("    movem.l %a0@, #0x0078   | move next 16 bytes to registers");
asm("    adda.l #16, %a0");
asm("   add.l %d3,%d1");
asm("   addx.l %d4,%d1");
asm("   addx.l %d5,%d1");
asm("   addx.l %d6,%d1");
asm("   addx.l %d2,%d1  | Overflow condition");
asm("    sub.l #16,%d0");
asm("   cmpi.l #16, %d0");
asm("   bge Quad_Dword_Loop");
asm("    movem.l %sp@, #0x0078 | restore d3-d6 from stack");
asm("   adda.l #16, %sp");
asm(" Sum_Dword:");
asm("   andil #0xFFFC,%d0 /* Truncate to dword */");
asm("    beq Byte_Sum");
asm(" Sum_Loop:");
asm("   add.l %a0@+,%d1");
asm("   addx.l %d2,%d1");
asm("    subq.l #4,%d0");
asm("   bne Sum_Loop");
asm(" Byte_Sum:     ");
asm("   movel %sp@(12),%d0 /* count */");
asm("   btst #1,%d0");
asm("   beq NotWord");
asm("    move.w %a0@+,%a1");
asm("    addl %a1,%d1");
asm("   addx.l %d2,%d1");
asm(" NotWord:");
asm("   btst #0,%d0");
asm("   beq SumDone");
asm("    move.b %a0@+,%d0");
asm("    asl.l #8, %d0");
asm("   andil #0xFF00,%d0");
asm("    addl %d0,%d1");
asm("   addx.l %d2,%d1");
asm(" SumDone:");
asm(" /* At this point the sum is 32 bits wide in d1 */");
asm("   move.w %d1,%d0");
asm("   swap %d1");
asm("   andil #0xFFFF,%d0 /* Truncate to word */");
asm("   andil #0xFFFF,%d1 /* Truncate to word */");
asm("   add.l %d1,%d0");
asm(" /* At this point the sum is 16 bits wide with possible carry */");
asm("   btst #16,%d0");
asm("   beq SumFixed");
asm("   andil #0xFFFF,%d0 /* Truncate to word */");
asm("   addq.l #1, %d0 /* add carry */");
asm(" SumFixed:");
asm("   movel  %sp@+, %d2  |restore %d2");
asm("   notl %d0" );
asm("   andil #0xFFFF,%d0 /* Truncate to word */");
asm("   move.l %d0,%d1");
asm("   rts");
asm(" GetSumEnd:");

/* WORD GetSum20( PDWORD addr ) */
asm(".global GetSum20");
asm("GetSum20:");
asm("   moveal %sp@(4),%a0 /* Pointer to data */");
asm("   clr.l %d1");
asm("   clr.l %d0");
asm("   add.l %a0@+,%d0");
asm("      addx.l %d1,%d0");
asm("   add.l %a0@+,%d0");
asm("      addx.l %d1,%d0");
asm("   add.l %a0@+,%d0");
asm("      addx.l %d1,%d0");
asm("   add.l %a0@+,%d0");
asm("      addx.l %d1,%d0");
asm("   add.l %a0@+,%d0");
asm("      addx.l %d1,%d0");
asm(" /* At this point the sum is 32 bits wide in d1 */");
asm("   move.l %d0,%d1");
asm("   swap %d1");
asm("   andil #0xFFFF,%d1 /* Truncate to word */");
asm("   add.l %d1,%d0");
asm(" /* At this point the sum is 16 bits wide with possible carry */");
asm("   clr.l %d1");
asm("      addx.l %d1,%d0");
asm("   notl %d0" );
asm("   andil #0xFFFF,%d0 /* Truncate to word */");
asm("   move.l %d0,%d1");
asm("   rts");
asm(" GetSum20End:");

/*WORD GetSumHdr( PsudeoHeader &hdr, PWORD addr, WORD count )*/
asm(".global GetSumHdr");
asm("GetSumHdr:");
asm("   moveal %sp@(8),%a0 | Pointer to data ");
asm("   clr.l %d1");
asm("   movel %d2, %sp@- |save %d2");
asm("   clr.l %d2");
asm("   movel %sp@(16),%d0 | count ");
asm("   cmpi.l #48, %d0");
asm("   blt SumHdr_Dword");
asm("   suba.l #16, %sp");
asm("    movem.l #0x0078, %sp@  | save d3-d6 to stack");
asm(" SumHdr_Quad_Loop:");
asm("    movem.l %a0@, #0x0078   | move next 16 bytes to registers");
asm("    adda.l #16, %a0");
asm("   add.l %d3,%d1");
asm("   addx.l %d4,%d1");
asm("   addx.l %d5,%d1");
asm("   addx.l %d6,%d1");
asm("   addx.l %d2,%d1  | Overflow condition");
asm("    sub.l #16,%d0");
asm("   cmpi.l #16, %d0");
asm("   bge SumHdr_Quad_Loop");
asm("    movem.l %sp@, #0x0078 | restore d3-d6 from stack");
asm("   adda.l #16, %sp");
asm(" SumHdr_Dword:");
asm("   andil #0xFFFC,%d0  | Truncate to word ");
asm("   beq Byte_SumHdr");
asm(" SumHdr_Loop:");
asm("   add.l %a0@+,%d1");
asm("   addx.l %d2,%d1");
asm("    subq.l #4,%d0");
asm("   bne SumHdr_Loop");
asm(" Byte_SumHdr:     ");
asm("   movel %sp@(16),%d0  | count ");
asm("   btst #1,%d0");
asm("   beq NotWord_Hdr");
asm("    move.w %a0@+,%a1");
asm("    addl %a1,%d1");
asm("   addx.l %d2,%d1");
asm("NotWord_Hdr:");
asm("   btst #0,%d0");
asm("   beq SumHdrTime");
asm("    move.b %a0@+,%d0");
asm("    asl.l #8, %d0");
asm("   andil #0xFF00,%d0");
asm("    addl %d0,%d1");
asm("   addx.l %d2,%d1");
asm(" SumHdrTime:");
asm("    movel %sp@(8),%a0  | header pointer ");
asm("   add.l %a0@+,%d1");
asm("   addx.l %d2,%d1");
asm("   add.l %a0@+,%d1");
asm("   addx.l %d2,%d1");
asm("   add.l %a0@+,%d1");
asm("   addx.l %d2,%d1");
asm(" SumHdrDone:");
asm(" | At this point the sum is 32 bits wide in d1 ");
asm("   move.w %d1,%d0");
asm("   swap %d1");
asm("   andil #0xFFFF,%d0  | Truncate to word ");
asm("   andil #0xFFFF,%d1  | Truncate to word ");
asm("   add.l %d1,%d0");
asm(" | At this point the sum is 16 bits wide with possible carry ");
asm("   btst #16,%d0");
asm("   beq SumHdrFixed");
asm("   andil #0xFFFF,%d0  | Truncate to word ");
asm("   addq.l #1, %d0     | add carry ");
asm(" SumHdrFixed:");
asm("   movel  %sp@+, %d2  |restore %d2");
asm("   notl %d0" );
asm("   andil #0xFFFF,%d0  | Truncate to word ");
asm("   move.l %d0,%d1");
asm("   rts");
asm(" GetSumHdrEnd:");

#endif

#ifdef C_CSUM
/*---------------------------------------------**
** Calculate a checksum starting at addr, and  **
** going for count bytes,                      **
** This is from RFC1071                        **
**---------------------------------------------*/
WORD GetSum( PWORD addr, WORD count )
{
   /* Compute Internet Checksum for "count" bytes
    *         beginning at location "addr".
    */
   register DWORD sum = 0;

   while ( count > 1 )
   {
      /*  This is the inner loop */
      sum += *addr++;
      count -= 2;
   }

   /*  Add left-over byte, if any */
   if ( count > 0 )
   {
      sum += ( ( *( ( unsigned char * ) addr ) ) << 8 );
   }

   /*  Fold 32-bit sum to 16 bits */
   while ( sum >> 16 )
   {
      sum = ( sum & 0xffff ) + ( sum >> 16 );
   }

   return ( WORD ) ~sum;
}

WORD GetSum20( PDWORD addrdw )
{
    register DWORD sum = 0;
    register PWORD addr=(PWORD)addrdw;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum += *addr++;
    sum = ( sum & 0xffff ) + ( sum >> 16 );
    if( sum & 0x10000 ) sum = ((sum & 0xFFFF) + 1);
    return ( WORD ) ~sum;
}


WORD GetSumHdr( PsudeoHeader &hdr, PWORD addr, WORD count )
{
   register int sum = 0;

   while ( count > 1 )
   {
      //   This is the inner loop /
      sum += *addr++;
      count -= 2;
   }
   //  Add left-over byte, if any /
   if ( count > 0 )
   {
      sum += ( ( *( ( unsigned char * ) addr ) ) << 8 );
   }

   addr = ( PWORD ) & hdr;
   sum += *addr++;
   sum += *addr++; //srcip
   sum += *addr++;
   sum += *addr++; //dstip
   sum += *addr++; //proto
   sum += *addr++; //length
   ;
   //  Fold 32-bit sum to 16 bits //
   while ( sum >> 16 )
   {
      sum = ( sum & 0xffff ) + ( sum >> 16 );
   }

   return ( WORD ) ~sum;
}


extern "C"
{
   DWORD memcpysum( void *out, const void *in, size_t count );
}


#endif


/*This is a general purpose routine that adjusts for header options
and returns the packet data
*/
PBYTE GetData( PIPPKT pIp )
{
   if ( pIp->bVerHdrLen == 0x45 )
   {
      return pIp->DATA;
   }
   else
   {
      return pIp->DATA + ( ( ( pIp->bVerHdrLen & 0x0F ) * 4 ) - 20 );
   }
}



BOOL IsMyIp4( IPADDR4 ip )
{
   if ( GetInterfaceForMyAddress4( ip ) > 0 )
   {
      return TRUE;
   }
   return FALSE;
}




void FixHeaderAndSend( PoolPtr p, PIPPKT pIp )
{
   register PEFRAME pf;

   SNMPCOUNT( snmp_var_ipOutRequests );

   //Fix the header check sum
   pIp->hCSum = 0;
   pIp->wpktId = wGlobalPktId++;
   if(pIp->ipSrc.IsNull())
        pIp->ipSrc = GetSrcIp4( pIp->ipDst );

   pIp->hCSum = GetSum( ( PWORD ) pIp, ( pIp->bVerHdrLen & 0x0f ) * 4 );

   //Fix the frame size and ehternet fields
   pf = GetEframe( p );

   p->usedsize = HTONS( pIp->wLength ) + 14;
   pf->eType = EIP_TYPE;


   if ( IsMyIp4( pIp->ipDst ) || ( ( pIp->ipDst & 0xFF000000 ) == 0x7F000000 ) )
   {
      register PoolPtr p2 = GetBuffer();
      if ( p2 != NULL )
      {
         p2->usedsize = p->usedsize;
         p2->bBufferFlag = p->bBufferFlag;
         memcpy( p2->pData, p->pData, p2->usedsize );
         pf = GetEframe( p2 );
         p2->bBufferFlag &= ~( BS_PHY_BCAST | BS_IP_BCAST );
         p2->bInterfaceNumber = GetInterfaceForMyAddress4( pIp->ipDst );
         DoIPPacket( p2,pf,~GetSum( ( PWORD ) GetData( pIp ),HTONS( pIp->wLength ) - ( ( pIp->bVerHdrLen & 0x0f ) * 4 ) ) );
      }
      FreeBuffer( p );
   }
   else
   {
      RouteOut( pf, pIp->ipDst, p );
   }
}


void FixHeaderAndSendViaInterface( PoolPtr p, PIPPKT pIp, int interface )
{
   register PEFRAME pf;

   SNMPCOUNT( snmp_var_ipOutRequests );

   //Fix the header check sum
   pIp->hCSum = 0;
   pIp->wpktId = wGlobalPktId++;
   pIp->ipSrc = InterfaceIP( interface );

   pIp->hCSum = GetSum( ( PWORD ) pIp, ( pIp->bVerHdrLen & 0x0f ) * 4 );

   //Fix the frame size and ehternet fields
   pf = GetEframe( p );

   p->usedsize = HTONS( pIp->wLength ) + 14;
   pf->eType = EIP_TYPE;

   if ( IsMyIp4( pIp->ipDst ) || ( ( pIp->ipDst & 0xFF000000 ) == 0x7F000000 ) )
   {
      register PoolPtr p2 = GetBuffer();
      if ( p2 != NULL )
      {
         p2->usedsize = p->usedsize;
         p2->bBufferFlag = p->bBufferFlag;
         memcpy( p2->pData, p->pData, p2->usedsize );
         pf = GetEframe( p2 );
         p2->bBufferFlag &= ~( BS_PHY_BCAST | BS_IP_BCAST );
         p2->bInterfaceNumber = interface;
         DoIPPacket( p2,
                     pf,
                     ~GetSum( ( PWORD ) GetData( pIp ),
                              HTONS( pIp->wLength ) - ( ( pIp->bVerHdrLen & 0x0f ) * 4 ) ) );
      }
      FreeBuffer( p );
   }
   else
   {
      RouteOutVia( pf, pIp->ipDst, p, interface );
   }
}






buffer_list ip_errs;


void SaveError( PoolPtr p )
{
   ip_errs.InsertTail( p );
   if ( ip_errs.GetCount() >= MAX_IP_ERRS )
   {
      FreeBuffer( ip_errs.RemoveHead() );
   }
}


/*--------------------------------------------------------------**
** Process ICMP packets.                                        **
** They pass checksum and general IP tests if the get this far. **
**--------------------------------------------------------------*/
void process_icmp4( PoolPtr p, PIPPKT pIp, WORD csum )
{
   register PBYTE pData = GetData( pIp );
   register BYTE type = pData[0];
   /*BYTE code=pData[1]; Not used*/



   /*We have passed in checksum over the entire IP packet after the header */
   /*this is == the ICMP check sum */
   if ( ( p->bBufferFlag & ( BS_PHY_BCAST | BS_IP_BCAST ) ) != 0 )
   {
      FreeBuffer( p );
      return;
   }
   if ( csum == 0xFFFF ) //The checksum worked
   {
      switch ( type )
      {
        case ICMP_TYPE_ECHO_REQUEST:
          {
             //Change the type
             pData[0] = ICMP_TYPE_ECHO_REPLY;
             //Swap the Src and Dest
             pIp->ipDst = pIp->ipSrc;
             pIp->bTTL = bTTL_Default;
             pData[2] = 0;
             pData[3] = 0;
             *( ( PWORD ) ( &pData[2] ) ) = GetSum( ( PWORD ) pData,
                                                    HTONS( pIp->wLength ) -
                                                    ( ( pIp->bVerHdrLen & 0x0f ) * 4 ) );
             frames_tx_icmp++;
             SNMPCOUNT( snmp_var_icmpInEchos );
             SNMPCOUNT( snmp_var_icmpOutEchoReps );
             SNMPCOUNT( snmp_var_icmpOutMsgs );
             FixHeaderAndSendViaInterface( p, pIp, p->bInterfaceNumber );
             /* return; Warning 527: Unreachable code at token 'break' */
          }
          break;
        case ICMP_TYPE_ECHO_REPLY:
          if ( IcmpFifo )
          {
             SNMPCOUNT( snmp_var_icmpInEchoReps );
             OSPoolFifoPost( IcmpFifo,  p );
          }
          else
          {
             FreeBuffer( p );
          }
          return;

        case ICMP_TYPE_REDIRECT_MESSAGE:
          SNMPCOUNT( snmp_var_icmpInRedirects );
          SaveError( p );
          break;
        case ICMP_TYPE_DEST_UNREACHABLE:
          SNMPCOUNT( snmp_var_icmpInDestUnreachs ); // Inc number of "ICMP Dest Unreach" packets received ( by R.A)
          SaveError( p );
          break;
        case ICMP_TYPE_SOURCE_QUENCH:
          SNMPCOUNT( snmp_var_icmpInSrcQuenchs );
          /*lint -fallthrough */
        default:
          SaveError( p );
      }//Switch
   }//If Csum ==0
   else
   {
      //ICMP Check sum failed
      ip_last_err = csum;
      DBPRINT( DB_IP, "ICMP Error\r\n" );
      SNMPCOUNT( snmp_var_icmpInErrors );  // Inc number of ICMP packets received with bad checksum (by R.A)
      frames_ip_errors++;
      SaveError( p );
   }
}



void DoIPPacket( PoolPtr pp, PEFRAME pF, WORD csum )
{
   register PIPPKT pIP = GetIpPkt( pF );

   SNMPCOUNT( snmp_var_ipInReceives );

   int nifc = pp->bInterfaceNumber;

#if defined MULTIHOME || defined AUTOIP
   if (pIP->ipDst!=InterfaceIP(nifc))
        {
         pp->bInterfaceNumber=GetMultiHomeInterface4(pIP->ipDst,nifc);
        }
#endif



   //Quickly toss bad IP packets
   if ( ( pIP->bTTL == 0 ) ||
        ( ( HTONS( pIP->wLength ) > pp->usedsize ) ) ||
        ( ( !IsMyIp4( pIP->ipDst ) ) &&
          ( !IsBroadCast( pIP->ipDst, nifc ) &&
            ( ( pIP->ipDst & 0xFF000000 ) != 0x7F000000 ) ) ) )
   {
      // We think we want to toss the packet, but before we do that one more check
      if ( !( ( pMultiCastFunc ) && pIP->ipDst.IsMultiCast() ) )
      {
         {
            InterfaceBlock *pifb = GetInterFaceBlock( nifc );

            if ( pifb == NULL )
            {
               pifb = GetInterFaceBlock();
            }

            if ( ( ( !pifb->netIP.IsGlobalBroadCast()) && ( pifb->netIP.NotNull() ) )  //The ethernet IP has been sent.
                ||
                 ( pp->bBufferFlag & BS_PHY_BCAST ) //Or the packet was an Ethernet Broadcast.... then really toss it
               )
            {
               if ( pPromisciousPacketFunc )
               {
                  pPromisciousPacketFunc( pp );
               }
               else
               // If our IP address is not yet valid (either 255.255.255.255 or 0.0.0.0), and
               // the Ethernet frame is a broadcast ( FF FF FF FF FF FF), allow the packet to
               // be dispatched.
               if ( ( ( pifb->netIP.IsGlobalBroadCast()) || ( pifb->netIP.IsNull() ) ) && (pp->bBufferFlag & BS_PHY_BCAST))
               {//This could be a CISCO broken DHCP packet, dispatch it if we must....
                pp->pPoolNext = ( PoolPtr ) ( DWORD ) csum;
                OSPoolFifoPost( &ipRxFifo, pp );
               }
               else
               {
                  FreeBuffer( pp );
                  frames_ip_discard++;
                  SNMPCOUNT( snmp_var_ipInAddrErrors );
               }
               return;
            }
         }
      }
   }//Probably Toss the packet

   if ( pIP->wFlags_Frag & 0x3FFF )
   {
      frames_rx_fragment++;
      if (pFragFunc) pFragFunc(pp);
      else
      FreeBuffer( pp );
      return;
   }


   pp->pPoolNext = ( PoolPtr ) ( DWORD ) csum;
   OSPoolFifoPost( &ipRxFifo,  pp );
}




void InitializeCStack( IPADDR4 ipaddr, IPADDR4 mask, IPADDR4 gate, IPADDR4 dns )
{
   InitializeStack( ipaddr, mask, gate, dns );
}



/*---------------------------------------------------------**
** Setup the IP stack, and start the timer tick interrupt. **
**---------------------------------------------------------*/
void InitializeStack( IPADDR4 ipaddr, IPADDR4 mask, IPADDR4 gate, IPADDR4 dns )
{
   bTTL_Default = 60;
   wArpLifetime = 120;
   pPromisciousPacketFunc = NULL;



   InitBuffers();
   InitializeArp();
   InitTCP();

   OSFifoInit( &ipRxFifo );

   OSTaskCreatewName( IPTask,
                 ( void * ) 0,
                 ( void * ) &IPStk[IP_STK_SIZE],
                 ( void * ) IPStk,
                 IP_PRIO,"IP" );
   IcmpFifo = NULL;


   InitializeNetwork( &DoIPPacket, &processArp );

   InterfaceBlock *pifb = GetInterFaceBlock();

   if ( pifb != NULL )
   {
      if ( ipaddr )
      {
         pifb->netIP = ipaddr;
      }
      if ( mask )
      {
         pifb->netIpMask = mask;
      }
      if ( gate )
      {
         pifb->netIpGate = gate;
      }
      if ( dns )
      {
         pifb->netDNS = dns;
      }
   }

#ifdef IPV6
    InitIPv6();
#endif

   if ( ( pifb != NULL ) && ( !gConfigRec.m_q_boot ) )
   {
      iprintf( "Configured IP = " );
      ShowIP4( pifb->netIP );
      iprintf( "\nConfigured Mask = " );
      ShowIP4( pifb->netIpMask );
      iprintf( "\nMAC Address= " );
      ShowMac( &pifb->theMac );
      iprintf( "\n" );
   }
}



void KillStack()
{
   StopNetworks();   /*lint !e522 */
}


PIPPKT PrepareIPOutput( PoolPtr p, IPADDR4 to, WORD protocol )
{
   register PIPPKT pIp = GetIpPkt( p );

   pIp->bVerHdrLen = 0x45;
   pIp->bTos = 0;
   pIp->wFlags_Frag = 0;
   pIp->ipDst = to;
   pIp->bTTL = bTTL_Default;
   pIp->proto = protocol;
   return pIp;
}

void FixUpICMP( PIPPKT pIp, BYTE typ, BYTE code, PBYTE data, WORD datasiz )
{
   pIp->DATA[0] = typ;
   pIp->DATA[1] = code;
   pIp->DATA[2] = 0;
   pIp->DATA[3] = 0;
   memcpy( ( pIp->DATA ) + 4, data, datasiz );  /*lint !e419 */

   WORD cs = GetSum( ( PWORD ) pIp->DATA, 4 + datasiz );

   pIp->wLength = HTONS( 20 + 4 + datasiz );

   *( PWORD ) ( pIp->DATA + 2 ) = cs;
}





/* This structure will never be created on its own
It is just a pattern of storage to be overlayed over the data
section of an IP packet.
It helps place and keep track of the data in an ICMP ping packet */

/*
 * ICMP echo request ("ping")
 *
 *    typ            - Type ICMP_TYPE_ECHO_REQUEST (8)
 *    code           - Code ICMP_CODE_ECHO_REQUEST (0)
 *    csum           - Checksum data and header with a zero here for calculation
 *    id             - ID from requestor
 *    seq            - Sequence value
 *    extra_data     - Data can be longer than ICMP_PING_DATA_SIZE bytes, just a placeholder
 *
 *    Notes:
 *       Defined in Internet Protocol Suite RFC 792
 *
 *       This structure will never be created on its own it is just a pattern of
 *       storage to be overlayed over the data section of an IP packet.
 *       It helps place and keep track of the data in an ICMP ping packet
 *
 */
struct ICMP_PING_REQUEST
{
   BYTE typ;
   BYTE code;
   WORD csum;
   WORD id;
   WORD seq;
   BYTE extra_data[ ICMP_PING_DATA_SIZE ];
};

/*
 *******************************************************************************
 *
 * Sends a "ping" packet and waits the specified timeout for a response.
 *
 *    SendPing does not wait for a response.
 *
 *******************************************************************************
 */
void SendPing4( IPADDR4 to, WORD id, WORD seq, int interface )
{
   PoolPtr p = GetBuffer();
   if ( p == 0 )
   {
      return;
   }

   PIPPKT pIp = PrepareIPOutput( p, to, IP_PROTOCOL_ICMP );
   pIp->ipSrc = 0;

   ICMP_PING_REQUEST *preq = (ICMP_PING_REQUEST*)pIp->DATA;

   preq->typ = ICMP_TYPE_ECHO_REQUEST;
   preq->code = ICMP_CODE_ECHO_REQUEST;
   preq->csum = 0 ;
   preq->id = HTONS( id );
   preq->seq = HTONS( seq );
   for ( int i = 0; i < ICMP_PING_DATA_SIZE; i++ )
   {
      preq->extra_data[i] = ( 'a' + i );
   }

   WORD cs = GetSum( (PWORD)pIp->DATA, sizeof( ICMP_PING_REQUEST ) );

   pIp->wLength = HTONS( ( 20 + sizeof( ICMP_PING_REQUEST ) ) );

   preq->csum = cs;

   SNMPCOUNT( snmp_var_icmpOutMsgs );   // Inc number of ICMP packets sent (by R.A)
   SNMPCOUNT( snmp_var_icmpOutEchos );  // Inc number of "ICMP Echo" packets sent (by R.A)

   if ( interface == -1 )
   {
      FixHeaderAndSend( p, pIp );
   }
   else
   {
      FixHeaderAndSendViaInterface( p, pIp, interface );
   }

   return;
}




void SendICMPError( PoolPtr pBadPacket, BYTE typ, BYTE code )
{
   register PoolPtr p = GetBuffer();
   if ( p == 0 )
   {
      return;
   }
   PIPPKT pIpsrc = GetIpPkt( pBadPacket );
   PDWORD lpdw = ( PDWORD ) pIpsrc;
   lpdw--;
   *lpdw = 0;

   PIPPKT pIp = PrepareIPOutput( p, pIpsrc->ipSrc, IP_PROTOCOL_ICMP );
   pIp->ipSrc = 0;

   FixUpICMP( pIp, typ, code, ( PBYTE ) lpdw, 32 );

   SNMPCOUNT( snmp_var_icmpOutDestUnreachs ); // Inc number of "ICMP Dest Unreach" packets sent ( FIXED by R.A)
   SNMPCOUNT( snmp_var_icmpOutMsgs );
   FixHeaderAndSend( p, pIp );
}

/*
 *******************************************************************************
 *
 * Sends a "ping" packet and waits the specified timeout for a response.
 *
 *    Ping uses the primary network interface (usually the ethernet)
 *    PingViaInterface allows the selection of network interface
 *
 *******************************************************************************
 */
int Ping4( IPADDR4 to, WORD id, WORD seq, WORD wait )
{
   return Ping4ViaInterface( to, id, seq, wait, -1 );
}


int Ping4ViaInterface( IPADDR4 to, WORD id, WORD seq, WORD wait, int interface )
{
   OS_FIFO myIcmpFifo;  //A FIFO structure to use for ICMP
   int ticksWaited = -1;

   OSFifoInit( &myIcmpFifo ); //Initialize it
   OS_FIFO *oldfifo = IcmpFifo; //Save the old one
   IcmpFifo = &myIcmpFifo; //Setup the new one

   int start = (int)TimeTick;
   SendPing4( to, id, seq, interface );
   //Now wait for a response
   pool_buffer *pb = OSPoolFifoPend( &myIcmpFifo, wait );
   IcmpFifo = oldfifo;
   //Record the completion time
   start = (int)TimeTick - start;

   if ( pb != NULL )
   {
      /* Get the data portion from the IP packet */
      ICMP_PING_REQUEST * preq =  (ICMP_PING_REQUEST *)GetData( GetIpPkt( pb ) );
      if (  ( preq->typ == ICMP_TYPE_ECHO_REPLY ) &&
            ( preq->code == ICMP_CODE_ECHO_REPLY) &&
            ( preq->id  == HTONS( id ) ) &&
            ( preq->seq == HTONS( seq ) ) )
      {
         ticksWaited = start;
         for ( int i = 0; i < ICMP_PING_DATA_SIZE; i++ )
         {
            if ( preq->extra_data[ i ] != ( 'a' + i ) )
            {
               ticksWaited = -1;
               break;
            }
         }
      }
      FreeBuffer( pb );
   }

   /* Purge fifo and free missed buffers if necessary */
   do
   {
      pb = OSPoolFifoPendNoWait( &myIcmpFifo );
      if ( pb == NULL )
      {
         break;
      }
      FreeBuffer( pb );

   } while( TRUE ); /*lint !e506 Constant value Boolean OK */

   return ticksWaited;
}


/**************************************Extra diagnostic functions ******************
void ShowIPBuffer( PoolPtr p )
{
   PEFRAME pF = GetEframe( p );
   if ( ( pF->eType != EIP_TYPE ) && ( pF->eType != PPP_TYPE ) )
   {
      iprintf( "Not an IP buffer \n" );
      ShowBuffer( p );
      return;
   }
   else
   {
      //IS an IP packet
      PIPPKT pIP = GetIpPkt( pF );

      switch ( pIP->proto )
      {
        case IP_PROTOCOL_ICMP:
          switch ( pIP->DATA[0] )
          {
            case 0:
              iprintf( "ICMP Echo Reply " ); break;
            case 3:
              iprintf( "ICMP Destination Unreachable code = %d", ( int ) pIP->DATA[1] ); break;
            case 4:
              iprintf( "ICMP Source Quench " ); break;
            case 5:
              iprintf( "ICMP Redirect = %d", ( int ) pIP->DATA[1] ); break;
            case 8:
              iprintf( "ICMP Echo Request " ); break;
            case 9:
              iprintf( "ICMP Router Advertisement " ); break;
            case 10:
              iprintf( "ICMP Router Solicitation " ); break;
            case 11:
              iprintf( "Time Exceeded Code = %d", ( int ) pIP->DATA[1] ); break;
            case 12:
              iprintf( "ICMP Parameter Problem " ); break;
            case 13:
              iprintf( "ICMP Timestamp Request " ); break;
            case 14:
              iprintf( "ICMP Timestamp Reply " ); break;
            case 17:
              iprintf( "ICMP Address Mask Request " ); break;
            case 18:
              iprintf( "ICMP Address MAsk Reply " ); break;
            default:
              iprintf( "ICMP Type Uknown " ); break;
          }
          break;
        case IP_PROTOCOL_UDP:
          {
             PUDPPKT pUdp = GetUdpPkt( pIP );
             iprintf( "UDP Packet\nSrc Port 0x%04x\nDest Port 0x%04x\nLength %d\nCheckSum %04X\n",
                      HTONS( pUdp->srcPort ),
                      HTONS( pUdp->dstPort ),
                      HTONS( pUdp->UdpLen ),
                      HTONS( pUdp->UdpCSum ) );
             ShowData( pUdp->DATA, HTONS( pUdp->UdpLen ) - 8 );
#ifdef _FULL_DISCLOSURE
             PsudeoHeader hdr;
             hdr.srcip = pIP->ipSrc;
             hdr.dstip = pIP->ipDst;
             hdr.proto = HTONS( 17 );
             hdr.len = pUdp->UdpLen;
             WORD csum = GetSumHdr( hdr, ( PWORD ) ( pUdp ), HTONS( pUdp->UdpLen ) );
             iprintf( "Calced Csum= %04X\n", csum );
             iprintf( "\nFull packet:\n" );
             ShowData( pF->pData, HTONS( pUdp->UdpLen ) + 48 );
#endif
          }

          break;
        case IP_PROTOCOL_IGMP:
          iprintf( "IGMP Packet\n" );
          ShowData( GetData( pIP ), 8 );
          ShowBuffer( p );
          iprintf( "\n" );
          break;

        case IP_PROTOCOL_TCP:
          iprintf( "TCP Packet\n" );
          DumpTcpPacket( pIP );
          break;

        default:
          iprintf( "Unknown IP proto = %d ", pIP->proto );
          ShowBuffer( p );
      }
      iprintf( "From " );
      ShowIP4( pIP->ipSrc );
      iprintf( " To " );
      ShowIP4( pIP->ipDst );
      iprintf( "\n" );
   }//Was an IP packet
}

*******************************************Extra diagnostics ***********************/




DWORD next_arp_time FAST_IP_VAR;
long delay FAST_IP_VAR;

void IPTask( void *p )
{
   next_arp_time = TimeTick + 10;

   while ( 1 )
   {
      delay = (long)( next_arp_time - TimeTick );

      if ( delay <= 0 )
      {
         AgeArp();
         next_arp_time = TimeTick + TICKS_PER_SECOND;
         delay = TICKS_PER_SECOND;
         if ( pDHCPTick )
         {
            pDHCPTick();
         }
         if ( pMultiTick )
         {
            pMultiTick();
         }
      }
#ifdef AUTOIP
      if ( pAutoIPTick && AutoIPClient::getNextTick() <= TimeTick )
      {
          pAutoIPTick();
      }
#endif

#ifdef IPV6
        if (pIpv6Tick)
        {
         pIpv6Tick();
        }

#endif



      register pool_buffer *pp = OSPoolFifoPend( &ipRxFifo, delay );

      if ( pp != NULL )
      {
         //We have a packet!
    	  i50msCountEthernet = 0;

         WORD csum = ( WORD ) ( DWORD ) pp->pPoolNext;
         register PEFRAME pF = GetEframe( pp );
         register PIPPKT pIP = GetIpPkt( pF );
         int nifc = pp->bInterfaceNumber;

         SNMPCOUNT( snmp_var_ipInDelivers );

         if ( IsBroadCast( pIP->ipDst, nifc ) )
         {
            pp->bBufferFlag |= BS_IP_BCAST;
         }

         if ( ( pMultiCastFunc ) && pIP->ipDst.IsMultiCast( ) )
         {
            /* Is Multicast */
            pMultiCastFunc( pp, csum );
            /* Is Multicast */
         }
         else
         {
            /* Not Multicast */

#ifdef NO_SYMETRIC_ROUTING
            if ( IsLocal4(pIP->ipSrc,nifc) )
               AddCheckArp( &( pF->src_addr ), pIP->ipSrc, nifc );
#else
            AddCheckArp( &( pF->src_addr ), pIP->ipSrc, nifc );
#endif
            switch ( pIP->proto )
            {
              case IP_PROTOCOL_ICMP:
                frames_rx_icmp++;
                SNMPCOUNT( snmp_var_icmpInMsgs );
                process_icmp4( pp, pIP, csum );

                break;
              case IP_PROTOCOL_UDP:
                frames_rx_udp++;
                process_udp4( pp, pIP, csum );

                break;
              case IP_PROTOCOL_TCP:
                frames_rx_tcp++;
                process_tcp4( pp, pIP, csum );
                break;
              default:
                ip_last_err = pIP->proto;
                DBPRINT( DB_IP, "Proto Error Error\r\n" );
                frames_ip_errors++;
                SNMPCOUNT( snmp_var_ipInUnknownProtos );
                SNMPDEC( snmp_var_ipInDelivers );
                FreeBuffer( pp );
            }//Switch on packet type
         }/* Not Multicast */
      } //We have a packet
   }//While (1)
}//Task Function end

/*
 ******************************************************************************
 *
 * Network Interface Controller (NIC) Stack Process Routines
 *
 ******************************************************************************
 */
/*
 ******************************************************************************

   Validates and submits Ethernet II style packets to the network stack.

   Parameters:
      poolPtr           - Pool packet containing the packet
      packetSizeInBytes - Size of packet as received from the interface.

   Return:
      None

   Notes:
      Must be Ethernet II style or it will be discarded.
      Packet interface (bInterfaceNumber) must be set to the interface number.
      Time in (dwTime, dwTimeFraction) must be set by module function.

 ******************************************************************************
 */
void IpProcessEthernetPacket( PoolPtr poolPtr, WORD packetSizeInBytes )
{
   register EthernetFrameHeader* frameHeaderPtr = NULL;
   register IpHeaderIPv4* ipHeaderPtr = NULL;
   register PBYTE datagramPtr = NULL;
   WORD optionDwordCount = 0;
   WORD headerLengthInBytes = 0;
   DWORD checkSum = 0;
   BOOL discardPacket = FALSE;


   /* Counters */
   frames_rx++;

   do
   {

      /* Packet sanity */
      if ( ( poolPtr == NULL ) ||
            ( packetSizeInBytes < sizeof( EthernetFrameHeader ) ) )
      {
         discardPacket = TRUE;
         break;
      }

      /* Packet auditing and controls */
      ChangeOwner( poolPtr );
      poolPtr->bBufferFlag = 0;


      /* Analyze ethernet frame header */
      frameHeaderPtr = ( EthernetFrameHeader* )&( poolPtr->pData[ 0 ] );

#ifdef IP_DEBUG

      if ( frameHeaderPtr == NULL )
      {
         /* Unlikely */
         discardPacket = TRUE;
         break;
      }

#endif /* IP_DEBUG */

      /* Process by type */
      switch ( frameHeaderPtr->etherType )
      {
         case ETHERNET_ETHERTYPE_IPv4:
            DBPRINT( DB_ETHER, "I" );
            ipHeaderPtr = ( IpHeaderIPv4* )( ( PBYTE )&( poolPtr->pData[ 0 ] ) +
                 sizeof( EthernetFrameHeader ) );

            if (
#ifdef IP_DEBUG
                  ( ipHeaderPtr == NULL ) ||
#endif /* IP_DEBUG */
                  ( ( ipHeaderPtr->versionNLength & IP_VERSION_MASK ) !=
                  IP_VERSION_IPv4 ) )
            {
               /* Bad packet */
               discardPacket = TRUE;
               break;
            }

            /* Datagram and count */
            headerLengthInBytes = sizeof( IpHeaderIPv4 );
            datagramPtr = ( PBYTE )ipHeaderPtr;

            if ( ( ipHeaderPtr->versionNLength & IP_HEADER_LENGTH_MASK ) !=
                  IP_HEADER_LENGTH_NO_OPTIONS )
            {
               optionDwordCount = ( WORD )( ipHeaderPtr->versionNLength &
                  IP_HEADER_LENGTH_MASK );
               optionDwordCount -= IP_HEADER_LENGTH_NO_OPTIONS;
               headerLengthInBytes += optionDwordCount * sizeof( DWORD );
            }

            /* Checksum data after IP header */
            datagramPtr += headerLengthInBytes;
            checkSum = GetSum( ( PWORD )datagramPtr,
                  ( ipHeaderPtr->totalLength - headerLengthInBytes ) );

            /* Adjust packet size */
            ChangeOwner( poolPtr );
            poolPtr->usedsize = sizeof( EthernetFrameHeader ) +
                  ipHeaderPtr->totalLength;

            /* Submit to stack */
#ifdef IP_DEBUG
            if ( pPacketfunc != NULL )
            {
               pPacketfunc( poolPtr, ( PEFRAME )frameHeaderPtr,
                     ( WORD )( ~checkSum ) );
            }
            else
            {
               /* Missing stack processor */
               discardPacket = TRUE;
            }
#else
            pPacketfunc( poolPtr, ( PEFRAME )frameHeaderPtr,
                  ( WORD )( ~checkSum ) );
#endif /* IP_DEBUG */
            break;

         case ETHERNET_ETHERTYPE_ARP:
            DBPRINT( DB_ETHER, "A" );

            /* Pad to 16 bit boundary */
            poolPtr->usedsize = ( ( ( packetSizeInBytes % 2 ) == 0 ) ?
                  packetSizeInBytes : ( packetSizeInBytes + 1 ) );


            /* Submit to stack */
            ChangeOwner( poolPtr );
#ifdef IP_DEBUG
            if ( pArpFunc != NULL )
            {
               pArpFunc( poolPtr, ( PEFRAME )frameHeaderPtr );
            }
            else
            {
               /* Missing stack processor */
               discardPacket = TRUE;
            }
#else
            pArpFunc( poolPtr, ( PEFRAME )frameHeaderPtr );
#endif /* IP_DEBUG */
            break;

         default:
            /* Unknown */
            discardPacket = TRUE;
            break;
      }

      /* Only once */
      break;

   } while ( TRUE ); /*lint !e506 Constant value Boolean */

   /* Packet controls */


   /* Need to discard? */
   if ( discardPacket == TRUE )
   {
      frames_rx_err++;
      FreeBuffer( poolPtr );
      poolPtr = NULL;
   }

   return;
}

WORD BspGetTickFraction();

WORD  GetEphemeralPort()
{
static WORD ephemeralport;
WORD rv;
 if ( ephemeralport < 0x8000 )
 {
     ephemeralport = 0x8000 + TimeTick+BspGetTickFraction();
     if ( ephemeralport < 0x8000 )
     {
         ephemeralport+= 0x8000;
     }
 }
 rv = ephemeralport++;
 if ( ephemeralport == 0 )
 {
     ephemeralport = 0x8000;
 }

return rv;
}
